"""A simple linear kinetic model."""

from jax import config
from jax import numpy as jnp

from enzax.kinetic_model import (
    KineticModel,
    KineticModelParameters,
    KineticModelStructure,
    UnparameterisedKineticModel,
)
from enzax.rate_equations import (
    AllostericIrreversibleMichaelisMenten,
    Drain,
    IrreversibleMichaelisMenten,
    ReversibleMichaelisMenten,
)


config.update("jax_enable_x64", True)

parameters = KineticModelParameters(
    log_kcat=jnp.log(
        jnp.array(
            [
                7.89577,  # MAT1
                19.9215,  # MAT3
                1.15777,  # METH-Gen
                10.5307,  # GNMT1
                234.284,  # AHC1
                1.77471,  # MS1
                13.7676,  # BHMT1
                7.02307,  # CBS1
                3.1654,  # MTHFR1
                0.264744,  # PROT1
            ]
        )
    ),
    log_enzyme=jnp.log(
        jnp.array(
            [
                0.000961712,  # MAT1
                0.00098812,  # MAT3
                0.00103396,  # METH-Gen
                0.000983692,  # GNMT1
                0.000977878,  # AHC1
                0.00105094,  # MS1
                0.000996603,  # BHMT1
                0.00134056,  # CBS1
                0.0010054,  # MTHFR1
                0.000995525,  # PROT1
            ]
        )
    ),
    log_drain=jnp.log(jnp.array([0.000850127])),
    dgf=jnp.array(
        [
            160.953,  # met-L
            -2263.31,  # atp
            -1055.95,  # pi
            -1943.8,  # ppi
            636.255,  # amet
            547.319,  # ahcys
            -161.373,  # gly
            -39.4573,  # sarcs
            44.2,  # hcys-L
            375.758,  # adn
            108.366,  # thf
            223.646,  # 5mthf
            198.009,  # mlthf
            173.094,  # glyb
            49.4547,  # dmgly
            -216.712,  # ser-L
            -2014.52,  # nadp
            -1948.58,  # nadph
            -46.4737,  # cyst-L
        ]
    ),
    log_km=jnp.log(
        jnp.array(
            [
                0.000106919,  # met-L MAT1
                0.00203015,  # atp MAT1
                0.00113258,  # met-L MAT3
                0.00236759,  # atp MAT3
                9.37e-06,  # amet METH-Gen
                0.000520015,  # amet GNMT1
                0.00253545,  # gly GNMT1
                2.32e-05,  # ahcys AHC1
                1.06e-05,  # hcys-L AHC1
                5.66e-06,  # adn AHC1
                1.71e-06,  # hcys-L MS1
                6.94e-05,  # 5mthf MS1
                1.98e-05,  # hcys-L BHMT1
                0.00845898,  # glyb BHMT1
                4.24e-05,  # hcys-L CBS1
                2.83e-06,  # ser-L CBS1
                8.08e-05,  # mlthf MTHFR1
                2.09e-05,  # nadph MTHFR1
                4.39e-05,  # met-L PROT1
            ]
        )
    ),
    log_ki=jnp.log(
        jnp.array(
            [
                0.000346704,  # MAT1
                5.56e-06,  # METH-Gen
                5.31e-05,  # GNMT1
            ]
        )
    ),
    log_conc_unbalanced=jnp.log(
        jnp.array(
            [
                # dataset1
                0.00131546,  # atp
                0.001,  # pi
                0.000500016,  # ppi
                0.00145177,  # gly
                1.00e-07,  # sarcs
                1.01e-06,  # adn
                2.24e-05,  # thf
                3.15e-06,  # mlthf
                0.00106758,  # glyb
                5.00e-05,  # dmgly
                0.0015873,  # ser-L
                1.22e-06,  # nadp
                0.000245139,  # nadph
                2.24e-06,  # cyst-L
            ]
        )
    ),
    temperature=jnp.array(298.15),
    log_transfer_constant=jnp.log(
        jnp.array(
            [
                0.107657,  # METAT
                131.207,  # GNMT
                1.03452,  # CBS
                0.392035,  # MTHFR
            ]
        )
    ),
    log_dissociation_constant=jnp.log(
        jnp.array(
            [
                0.000316641,  # amet MAT3
                0.00059999,  # met-L MAT3
                1.98e-05,  # amet GNMT1
                0.000228576,  # mlthf GNMT1
                9.30e-05,  # amet CBS1
                1.46e-05,  # amet MTHFR1
                2.45e-06,  # ahcys MTHFR1
            ]
        )
    ),
)
structure = KineticModelStructure(
    S=jnp.array(
        [
            [1, -1, -1, 0, 0, 0, 1, 1, 0, 0, -1],  # met-L b
            [0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0],  # atp
            [0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],  # pi
            [0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],  # ppi
            [0, 1, 1, -1, -1, 0, 0, 0, 0, 0, 0],  # amet b
            [0, 0, 0, 1, 1, -1, 0, 0, 0, 0, 0],  # ahcys b
            [0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0],  # gly
            [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],  # sarcs
            [0, 0, 0, 0, 0, 1, -1, -1, -1, 0, 0],  # hcys-L b
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],  # adn
            [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],  # thf
            [0, 0, 0, 0, 0, 0, -1, 0, 0, 1, 0],  # 5mthf b
            [0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0],  # mlthf
            [0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0],  # glyb
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],  # dmgly
            [0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0],  # ser-L
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],  # nadp
            [0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0],  # nadph
            [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],  # cyst-L
        ]
    ),
    water_stoichiometry=jnp.array(
        [
            0.0,  # met-L source
            -1.0,  # MAT1 METAT
            -1.0,  # MAT3 METAT
            0.0,  # METH Gen
            0.0,  # GNMT1
            -1.0,  # AHC
            0.0,  # MS
            0.0,  # BHMT
            1.0,  # CBS
            0.0,  # MTHFR
            0.0,  # PROT
        ]
    ),
    balanced_species=jnp.array(
        [
            0,  # met-L
            4,  # amet
            5,  # ahcys
            8,  # hcys-L
            11,  # 5mthf
        ]
    ),
    rate_to_enzyme_ix=[
        [],  # met-L source
        [0],  # MAT1 METAT
        [1],  # MAT3 METAT
        [2],  # METH Gen
        [3],  # GNMT1
        [4],  # AHC
        [5],  # MS
        [6],  # BHMT
        [7],  # CBS
        [8],  # MTHFR
        [9],  # PROT
    ],
    rate_to_km_ixs=[
        [],  # met-L source
        [0, 1],  # MAT1 METAT
        [2, 3],  # MAT3 METAT
        [4],  # METH Gen
        [5, 6],  # GNMT1
        [7, 8, 9],  # AHC
        [10, 11],  # MS
        [12, 13],  # BHMT
        [14, 15],  # CBS
        [16, 17],  # MTHFR
        [18],  # PROT
    ],
    species_to_metabolite_ix=jnp.arange(19),
    rate_to_subunits=jnp.array(
        [
            0,  # met-L source
            1,  # MAT1 METAT
            2,  # MAT3 METAT
            1,  # METH Gen
            4,  # GNMT1
            1,  # AHC
            1,  # MS
            1,  # BHMT
            2,  # CBS
            2,  # MTHFR
            1,  # PROT
        ]
    ),
    rate_to_tc_ix=[
        [],  # met-L source
        [],  # MAT1 METAT
        [0],  # MAT3 METAT
        [],  # METH Gen
        [1],  # GNMT1
        [],  # AHC
        [],  # MS
        [],  # BHMT
        [2],  # CBS
        [3],  # MTHFR
        [],  # PROT
    ],
    rate_to_dc_ixs_activation=[
        [],  # met-L source
        [],  # MAT1 METAT
        [0, 1],  # MAT3 METAT
        [],  # METH Gen
        [2],  # GNMT1
        [],  # AHC
        [],  # MS
        [],  # BHMT
        [4],  # CBS
        [6],  # MTHFR
        [],  # PROT
    ],
    rate_to_dc_ixs_inhibition=[
        [],  # met-L source
        [],  # MAT1 METAT
        [],  # MAT3 METAT
        [],  # METH Gen
        [3],  # GNMT1
        [],  # AHC
        [],  # MS
        [],  # BHMT
        [],  # CBS
        [5],  # MTHFR
        [],  # PROT
    ],
    rate_to_drain_ix=[
        [0],  # met-L source
        [],  # MAT1 METAT
        [],  # MAT3 METAT
        [],  # METH Gen
        [],  # GNMT1
        [],  # AHC
        [],  # MS
        [],  # BHMT
        [],  # CBS
        [],  # MTHFR
        [],  # PROT
    ],
    drain_sign=jnp.array([1.0]),
    dc_to_species_ix=jnp.array(
        [
            4,  # amet -> MAT3
            0,  # met-L -> MAT3
            4,  # amet -> GNMT1
            12,  # mlthf -| GNMT1
            4,  # amet -> CBS1
            4,  # amet -| MTHFR1
            5,  # ahcys -> MTHFR1
        ]
    ),
    ki_to_species_ix=jnp.array(
        [
            4,  # amet -| MAT1
            5,  # ahcys -| METH-GEN
            5,  # ahcys -| GNMT1
        ]
    ),
    rate_to_ki_ixs=[
        [],  # met-L source
        [0],  # MAT1 METAT
        [],  # MAT3 METAT
        [1],  # METH Gen
        [2],  # GNMT1
        [],  # AHC
        [],  # MS
        [],  # BHMT
        [],  # CBS
        [],  # MTHFR
        [],  # PROT
    ],
)
unparameterised_model = UnparameterisedKineticModel(
    structure,
    [
        Drain,  # met-L source
        IrreversibleMichaelisMenten,  # MAT1
        AllostericIrreversibleMichaelisMenten,  # MAT3
        IrreversibleMichaelisMenten,  # METH
        AllostericIrreversibleMichaelisMenten,  # GNMT1
        ReversibleMichaelisMenten,  # AHC
        IrreversibleMichaelisMenten,  # MS
        IrreversibleMichaelisMenten,  # BHMT
        AllostericIrreversibleMichaelisMenten,  # CBS
        AllostericIrreversibleMichaelisMenten,  # MTHFR
        IrreversibleMichaelisMenten,  # PROT
    ],
)
model = KineticModel(parameters, unparameterised_model)
steady_state = jnp.array(
    [
        4.233000e-05,  # met-L
        3.099670e-05,  # amet
        2.170170e-07,  # ahcys
        3.521780e-06,  # hcys
        6.534400e-06,  # 5mthf
    ]
)



from jax import numpy as jnp
import pytest

from enzax.examples import linear, methionine


@pytest.mark.parametrize(
    ["model", "steady_state"],
    [
        (methionine.model, methionine.steady_state),
        (linear.model, linear.steady_state),
    ],
)
def test_dcdt(model, steady_state):
    """Test for near-zero dcdt at a known steady state."""

    dcdt = model.dcdt(0, steady_state)
    zero = jnp.full((len(steady_state),), 0.0)
    assert jnp.isclose(dcdt, zero).all()



"""Code for doing mcmc on the parameters of a steady state problem."""

import functools

import blackjax
import chex
import equinox as eqx
import jax
import jax.numpy as jnp
from jax.scipy.stats import norm

from enzax.examples import methionine
from enzax.kinetic_model import (
    KineticModelParameters,
    KineticModel,
    UnparameterisedKineticModel,
)
from enzax.rate_equations import (
    AllostericReversibleMichaelisMenten,
    ReversibleMichaelisMenten,
)
from enzax.steady_state_problem import solve
from jaxtyping import Array, Float, ScalarLike

SEED = 1234
jax.config.update("jax_enable_x64", True)
jax.config.update("jax_debug_nans", True)


@chex.dataclass
class ObservationSet:
    conc: Float[Array, " m"]
    flux: ScalarLike
    enzyme: Float[Array, " e"]
    conc_scale: ScalarLike
    flux_scale: ScalarLike
    enzyme_scale: ScalarLike


@chex.dataclass
class PriorSet:
    log_kcat: Float[Array, "2 n"]
    log_enzyme: Float[Array, "2 n"]
    dgf: Float[Array, "2 n_metabolite"]
    log_km: Float[Array, "2 n_km"]
    log_conc_unbalanced: Float[Array, "2 n_unbalanced"]
    temperature: Float[Array, "2"]
    log_ki: Float[Array, " n_ki"]
    log_transfer_constant: Float[Array, "2 n_allosteric_enzyme"]
    log_dissociation_constant: Float[Array, "2 n_allosteric_effector"]


def ind_normal_prior_logdensity(param, prior):
    return norm.logpdf(param, loc=prior[0], scale=prior[1]).sum()


@eqx.filter_jit
def posterior_logdensity_fn(
    parameters: KineticModelParameters,
    unparameterised_model: UnparameterisedKineticModel,
    obs: ObservationSet,
    prior: PriorSet,
    guess: Float[Array, " n_balanced"],
):
    model = KineticModel(parameters, unparameterised_model)
    steady = solve(parameters, unparameterised_model, guess)
    flux = model.flux(steady)
    conc = jnp.zeros(model.structure.S.shape[0])
    conc = conc.at[model.structure.balanced_species].set(steady)
    conc = conc.at[model.structure.unbalanced_species].set(
        jnp.exp(parameters.log_conc_unbalanced)
    )
    likelihood_logdensity = (
        norm.logpdf(jnp.log(obs.conc), jnp.log(conc), obs.conc_scale).sum()
        + norm.logpdf(obs.flux, flux[0], obs.flux_scale).sum()
        + norm.logpdf(
            jnp.log(obs.enzyme), parameters.log_enzyme, obs.enzyme_scale
        ).sum()
    )
    prior_logdensity = (
        ind_normal_prior_logdensity(parameters.log_kcat, prior.log_kcat)
        + ind_normal_prior_logdensity(parameters.log_enzyme, prior.log_enzyme)
        + ind_normal_prior_logdensity(parameters.dgf, prior.dgf)
        + ind_normal_prior_logdensity(parameters.log_km, prior.log_km)
        + ind_normal_prior_logdensity(
            parameters.log_conc_unbalanced, prior.log_conc_unbalanced
        )
        + ind_normal_prior_logdensity(parameters.temperature, prior.temperature)
        + ind_normal_prior_logdensity(parameters.log_ki, prior.log_ki)
        + ind_normal_prior_logdensity(
            parameters.log_transfer_constant, prior.log_transfer_constant
        )
        + ind_normal_prior_logdensity(
            parameters.log_dissociation_constant,
            prior.log_dissociation_constant,
        )
    )
    return prior_logdensity + likelihood_logdensity


@functools.partial(jax.jit, static_argnames=["kernel", "num_samples"])
def inference_loop(rng_key, kernel, initial_state, num_samples):
    def one_step(state, rng_key):
        state, _ = kernel(rng_key, state)
        return state, state

    keys = jax.random.split(rng_key, num_samples)
    _, states = jax.lax.scan(one_step, initial_state, keys)

    return states


@eqx.filter_jit
def sample(logdensity_fn, rng_key, init_parameters):
    warmup = blackjax.window_adaptation(
        blackjax.nuts,
        logdensity_fn,
        progress_bar=True,
        initial_step_size=0.0001,
        max_num_doublings=10,
        is_mass_matrix_diagonal=False,
        target_acceptance_rate=0.95,
    )
    rng_key, warmup_key = jax.random.split(rng_key)
    (initial_state, tuned_parameters), _ = warmup.run(
        warmup_key,
        init_parameters,
        num_steps=1000,  # type: ignore
    )
    rng_key, sample_key = jax.random.split(rng_key)
    nuts_kernel = blackjax.nuts(logdensity_fn, **tuned_parameters).step
    states = inference_loop(
        sample_key,
        kernel=nuts_kernel,
        initial_state=initial_state,
        num_samples=200,
    )
    return states


def ind_prior_from_truth(truth, sd):
    return jnp.vstack((truth, jnp.full(truth.shape, sd)))


def main():
    """Demonstrate the functionality of the mcmc module."""
    true_parameters = methionine.parameters
    unparameterised_model = UnparameterisedKineticModel(
        structure=methionine.structure,
        rate_equation_classes=[
            AllostericReversibleMichaelisMenten,
            AllostericReversibleMichaelisMenten,
            ReversibleMichaelisMenten,
        ],
    )
    default_state_guess = jnp.array([0.1, 0.1])
    true_model = KineticModel(
        parameters=true_parameters, unparameterised_model=unparameterised_model
    )
    true_states = solve(
        true_parameters, unparameterised_model, default_state_guess
    )
    prior = PriorSet(
        log_kcat=ind_prior_from_truth(true_parameters.log_kcat, 0.1),
        log_enzyme=ind_prior_from_truth(true_parameters.log_enzyme, 0.1),
        dgf=ind_prior_from_truth(true_parameters.dgf, 0.1),
        log_km=ind_prior_from_truth(true_parameters.log_km, 0.1),
        log_conc_unbalanced=ind_prior_from_truth(
            true_parameters.log_conc_unbalanced, 0.1
        ),
        temperature=ind_prior_from_truth(true_parameters.temperature, 0.1),
        log_ki=ind_prior_from_truth(true_parameters.log_ki, 0.1),
        log_transfer_constant=ind_prior_from_truth(
            true_parameters.log_transfer_constant, 0.1
        ),
        log_dissociation_constant=ind_prior_from_truth(
            true_parameters.log_dissociation_constant, 0.1
        ),
    )
    # get true concentration
    true_conc = jnp.zeros(methionine.structure.S.shape[0])
    true_conc = true_conc.at[methionine.structure.balanced_species].set(
        true_states
    )
    true_conc = true_conc.at[methionine.structure.unbalanced_species].set(
        jnp.exp(true_parameters.log_conc_unbalanced)
    )
    # get true flux
    true_flux = true_model.flux(true_states)
    # simulate observations
    error_conc = 0.03
    error_flux = 0.05
    error_enzyme = 0.03
    key = jax.random.key(SEED)
    obs_conc = jnp.exp(jnp.log(true_conc) + jax.random.normal(key) * error_conc)
    obs_enzyme = jnp.exp(
        true_parameters.log_enzyme + jax.random.normal(key) * error_enzyme
    )
    obs_flux = true_flux + jax.random.normal(key) * error_conc
    obs = ObservationSet(
        conc=obs_conc,
        flux=obs_flux,
        enzyme=obs_enzyme,
        conc_scale=error_conc,
        flux_scale=error_flux,
        enzyme_scale=error_enzyme,
    )
    log_M = functools.partial(
        posterior_logdensity_fn,
        obs=obs,
        prior=prior,
        unparameterised_model=unparameterised_model,
        guess=default_state_guess,
    )
    samples = sample(log_M, key, true_parameters)
    print("True parameter values vs posterior:")
    for param in true_parameters.__dataclass_fields__.keys():
        true_val = getattr(true_parameters, param)
        model_low = jnp.quantile(getattr(samples.position, param), 0.01, axis=0)
        model_high = jnp.quantile(
            getattr(samples.position, param), 0.99, axis=0
        )
        print(f" {param}:")
        print(f"  true value: {true_val}")
        print(f"  posterior 1%: {model_low}")
        print(f"  posterior 99%: {model_high}")


if __name__ == "__main__":
    main()



"""Module containing rate equations for enzyme-catalysed reactions."""

from abc import abstractmethod

from jax import numpy as jnp
from jaxtyping import Array, Float, Int, Scalar

from enzax.kinetic_model import (
    KineticModelParameters,
    KineticModelStructure,
    RateEquation,
)


class Drain(RateEquation):
    log_v: Scalar
    sign: Scalar

    def __init__(
        self,
        parameters: KineticModelParameters,
        structure: KineticModelStructure,
        ix: int,
    ):
        self.log_v = parameters.log_drain[
            jnp.array(structure.rate_to_drain_ix[ix][0])
        ]
        self.sign = structure.drain_sign[
            jnp.array(structure.rate_to_drain_ix[ix])[0]
        ]

    def __call__(self, conc: Float[Array, " n"]) -> Scalar:
        """Get flux of a drain reaction.

        :param conc: A 1D array of non-negative numbers representing concentrations of the species that the reaction produces and consumes.

        """
        return self.sign * jnp.exp(self.log_v)


class MichaelisMenten(RateEquation):
    """Abstract base class for Michaelis Menten rate equations.

    Subclasses need to implement the method free_enzyme_ratio (and also the __call__ method).

    """

    dgf: Float[Array, " n"]
    log_km: Float[Array, " n"]
    log_enzyme: Scalar
    log_kcat: Scalar
    log_ki: Float[Array, " n_ki"]
    temperature: Scalar
    stoich: Float[Array, " n"]
    ix_substrate: Int[Array, " n_substrate"]
    ix_ki_species: Int[Array, " n_ki"]
    water_stoichiometry: Scalar
    substrate_km_positions: Int[Array, " n_substrate"]
    substrate_reactant_positions: Int[Array, " n_substrate"]

    def __init__(
        self,
        parameters: KineticModelParameters,
        structure: KineticModelStructure,
        ix: int,
    ):
        ix_dgf = structure.species_to_metabolite_ix[
            structure.rate_to_reactants[ix]
        ]
        self.dgf = parameters.dgf[ix_dgf]
        self.log_km = parameters.log_km[structure.rate_to_km_ixs[ix]]
        self.log_enzyme = parameters.log_enzyme[
            structure.rate_to_enzyme_ix[ix][0]
        ]
        self.log_kcat = parameters.log_kcat[structure.rate_to_enzyme_ix[ix][0]]
        self.log_ki = parameters.log_ki[structure.rate_to_ki_ixs[ix]]
        self.temperature = parameters.temperature
        self.stoich = structure.rate_to_stoichs[ix]
        self.ix_substrate = structure.rate_to_substrates[ix]
        self.substrate_km_positions = jnp.arange(len(self.ix_substrate))
        self.substrate_reactant_positions = (
            structure.rate_to_substrate_reactant_positions[ix]
        )
        self.ix_ki_species = structure.ki_to_species_ix[
            structure.rate_to_ki_ixs[ix]
        ]
        self.water_stoichiometry = structure.water_stoichiometry[ix]

    @property
    def km(self):
        return jnp.exp(self.log_km)

    @property
    def kcat(self):
        return jnp.exp(self.log_kcat)

    @property
    def ki(self):
        return jnp.exp(self.log_ki)

    @property
    def enzyme(self):
        return jnp.exp(self.log_enzyme)

    def numerator(self, conc: Float[Array, " n"]) -> Scalar:
        """Get the product of each substrate's concentration over its km.
        This quantity is the numerator in a Michaelis Menten reaction's rate equation
        """
        return jnp.prod(
            (conc[self.ix_substrate] / self.km[self.substrate_km_positions])
        )

    @abstractmethod
    def free_enzyme_ratio(self, conc: Float[Array, " n"]) -> Scalar: ...


class IrreversibleMichaelisMenten(MichaelisMenten):
    def free_enzyme_ratio(self, conc: Float[Array, " n"]) -> Scalar:
        return 1.0 / (
            jnp.prod(
                (
                    (
                        conc[self.ix_substrate]
                        / self.km[self.substrate_km_positions]
                    )
                    + 1
                )
                ** jnp.abs(self.stoich[self.substrate_reactant_positions])
            )
            + jnp.sum(conc[self.ix_ki_species] / self.ki)
        )

    def __call__(self, conc: Float[Array, " n"]) -> Scalar:
        """Get flux of a reaction with irreversible Michaelis Menten kinetics.

        :param conc: A 1D array of non-negative numbers representing concentrations of the species that the reaction produces and consumes.

        """
        saturation: Scalar = self.numerator(conc) * self.free_enzyme_ratio(conc)
        out = self.kcat * self.enzyme * saturation
        return out


class ReversibleMichaelisMenten(MichaelisMenten):
    ix_product: Int[Array, " n_product"]
    ix_reactants: Int[Array, " n_reactant"]
    product_reactant_positions: Int[Array, " n_product"]

    def __init__(
        self,
        parameters: KineticModelParameters,
        structure: KineticModelStructure,
        ix: int,
    ):
        super().__init__(parameters, structure, ix)
        self.ix_product = structure.rate_to_products[ix]
        self.ix_reactants = structure.rate_to_reactants[ix]
        self.product_reactant_positions = (
            structure.rate_to_product_reactant_positions[ix]
        )

    def reversibility(
        self, conc: Float[Array, " n"], water_stoichiometry: Scalar
    ) -> Scalar:
        """Get the reversibility of a reaction.

        Hard coded water dgf is taken from <http://equilibrator.weizmann.ac.il/metabolite?compoundId=C00001>.

        """
        RT = self.temperature * 0.008314
        dgf_water = -150.9
        dgr = self.stoich @ self.dgf + water_stoichiometry * dgf_water
        quotient = self.stoich @ jnp.log(conc[self.ix_reactants])
        out = 1.0 - jnp.exp(((dgr + RT * quotient) / RT))
        return out

    def free_enzyme_ratio(self, conc: Float[Array, " n"]) -> Scalar:
        return 1.0 / (
            -1.0
            + jnp.prod(
                (
                    (
                        conc[self.ix_substrate]
                        / self.km[self.substrate_reactant_positions]
                    )
                    + 1.0
                )
                ** jnp.abs(self.stoich[self.substrate_reactant_positions])
            )
            + jnp.prod(
                (
                    (
                        conc[self.ix_product]
                        / self.km[self.product_reactant_positions]
                    )
                    + 1.0
                )
                ** jnp.abs(self.stoich[self.product_reactant_positions])
            )
            + jnp.sum(conc[self.ix_ki_species] / self.ki)
        )

    def __call__(self, conc: Float[Array, " n"]) -> Scalar:
        """Get flux of a reaction with reversible Michaelis Menten kinetics.

        :param conc: A 1D array of non-negative numbers representing concentrations of the species that the reaction produces and consumes.

        """
        saturation: Scalar = self.numerator(conc) * self.free_enzyme_ratio(conc)
        out = (
            self.reversibility(conc, self.water_stoichiometry)
            * self.kcat
            * self.enzyme
            * saturation
        )
        return out


class AllostericRateLaw(MichaelisMenten):
    """Mixin class providing the method allosteric_effect."""

    subunits: int
    species_activation: Int[Array, " n_activation"]
    species_inhibition: Int[Array, " n_inhibition"]
    tc: Scalar
    dc_activation: Float[Array, " n_activation"]
    dc_inhibition: Float[Array, " n_inhibition"]

    def __init__(self, parameters, structure, ix):
        ix_dc_activation = jnp.array(
            structure.rate_to_dc_ixs_activation[ix], dtype=jnp.int16
        )
        ix_dc_inhibition = jnp.array(
            structure.rate_to_dc_ixs_inhibition[ix], dtype=jnp.int16
        )
        super().__init__(parameters, structure, ix)
        self.subunits = structure.rate_to_subunits[ix]  # type: ignore
        self.species_activation = structure.dc_to_species_ix[ix_dc_activation]
        self.species_inhibition = structure.dc_to_species_ix[ix_dc_inhibition]
        self.tc = jnp.exp(
            parameters.log_transfer_constant[structure.rate_to_tc_ix[ix][0]]
        )
        self.dc_activation = jnp.exp(
            parameters.log_dissociation_constant[ix_dc_activation]
        )
        self.dc_inhibition = jnp.exp(
            parameters.log_dissociation_constant[ix_dc_inhibition]
        )

    def allosteric_effect(self, conc: Float[Array, " n"]) -> Scalar:
        qnum = 1 + jnp.sum(conc[self.species_inhibition] / self.dc_inhibition)
        qdenom = 1 + jnp.sum(conc[self.species_activation] / self.dc_activation)
        out = 1.0 / (
            1
            + self.tc
            * (self.free_enzyme_ratio(conc) * qnum / qdenom) ** self.subunits
        )
        return out


class AllostericIrreversibleMichaelisMenten(
    AllostericRateLaw, IrreversibleMichaelisMenten
):
    def __call__(self, conc: Float[Array, " n"]) -> Scalar:
        out = super().__call__(conc) * self.allosteric_effect(conc)
        return out


class AllostericReversibleMichaelisMenten(
    AllostericRateLaw, ReversibleMichaelisMenten
):
    def __call__(self, conc: Float[Array, " n"]) -> Scalar:
        return super().__call__(conc) * self.allosteric_effect(conc)



"""A simple linear kinetic model."""

from jax import config
from jax import numpy as jnp

from enzax.kinetic_model import (
    KineticModel,
    KineticModelParameters,
    KineticModelStructure,
    UnparameterisedKineticModel,
)
from enzax.rate_equations import (
    AllostericReversibleMichaelisMenten,
    ReversibleMichaelisMenten,
)

config.update("jax_enable_x64", True)

parameters = KineticModelParameters(
    log_kcat=jnp.array([-0.1, 0.0, 0.1]),
    log_enzyme=jnp.log(jnp.array([0.3, 0.2, 0.1])),
    dgf=jnp.array([-3, -1.0]),
    log_km=jnp.array([0.1, -0.2, 0.5, 0.0, -1.0, 0.5]),
    log_ki=jnp.array([1.0]),
    log_conc_unbalanced=jnp.log(jnp.array([0.5, 0.1])),
    temperature=jnp.array(310.0),
    log_transfer_constant=jnp.array([-0.2, 0.3]),
    log_dissociation_constant=jnp.array([-0.1, 0.2]),
    log_drain=jnp.array([]),
)
structure = KineticModelStructure(
    S=jnp.array([[-1, 0, 0], [1, -1, 0], [0, 1, -1], [0, 0, 1]]),
    water_stoichiometry=jnp.array([0, 0, 0]),
    balanced_species=jnp.array([1, 2]),
    rate_to_enzyme_ix=[[0], [1], [2]],
    rate_to_km_ixs=jnp.array([[0, 1], [2, 3], [4, 5]]),
    species_to_metabolite_ix=jnp.array([0, 0, 1, 1]),
    rate_to_subunits=jnp.array([1, 1, 1]),
    rate_to_tc_ix=[[0], [1], []],
    rate_to_dc_ixs_activation=[[0], [], []],
    rate_to_dc_ixs_inhibition=[[], [1], []],
    rate_to_drain_ix=[[], [], []],
    drain_sign=[],
    dc_to_species_ix=jnp.array([2, 1]),
    ki_to_species_ix=jnp.array([1]),
    rate_to_ki_ixs=[[], [0], []],
)
unparameterised_model = UnparameterisedKineticModel(
    structure,
    [
        AllostericReversibleMichaelisMenten,
        AllostericReversibleMichaelisMenten,
        ReversibleMichaelisMenten,
    ],
)
model = KineticModel(parameters, unparameterised_model)
steady_state = jnp.array([0.43658744, 0.12695706])



"""Given a structural kinetic model, a set of parameters and an initial guess, find the physiological steady state metabolite concentration and its parameter sensitivities."""

import time


import diffrax
import equinox as eqx
import jax
import jax.numpy as jnp
import lineax as lx
from jaxtyping import Array, Float
from enzax.examples import methionine
from enzax.kinetic_model import (
    KineticModel,
    KineticModelParameters,
    UnparameterisedKineticModel,
)

jax.config.update("jax_debug_nans", True)
jax.config.update("jax_enable_x64", True)


@eqx.filter_jit()
def solve(
    parameters: KineticModelParameters,
    unparameterised_model: UnparameterisedKineticModel,
    guess: Float[Array, " n"],
):
    model = KineticModel(parameters, unparameterised_model)
    term = diffrax.ODETerm(model.dcdt)
    solver = diffrax.Kvaerno5()
    t0 = 0
    t1 = 900
    dt0 = 0.000001
    max_steps = None
    controller = diffrax.PIDController(
        pcoeff=0.1, icoeff=0.3, rtol=1e-11, atol=1e-11
    )
    cond_fn = diffrax.steady_state_event()
    event = diffrax.Event(cond_fn)
    adjoint = diffrax.ImplicitAdjoint(
        linear_solver=lx.AutoLinearSolver(well_posed=False)
    )
    sol = diffrax.diffeqsolve(
        term,
        solver,
        t0,
        t1,
        dt0,
        guess,
        args=model,
        max_steps=max_steps,
        stepsize_controller=controller,
        event=event,
        adjoint=adjoint,
        # progress_meter=diffrax.TextProgressMeter(minimum_increase=0.001),
    )
    return sol.ys[0]


def main():
    """Function for testing the steady state solver."""
    # guesses
    bad_guess = jnp.full((5,), 0.01)
    good_guess = jnp.array(
        [
            4.233000e-05,  # met-L
            3.099670e-05,  # amet
            2.170170e-07,  # ahcys
            3.521780e-06,  # hcys
            6.534400e-06,  # 5mthf
        ]
    )
    model = methionine.model
    # solve once for jitting
    solve(methionine.parameters, methionine.unparameterised_model, good_guess)
    jax.jacrev(solve)(
        methionine.parameters, methionine.unparameterised_model, good_guess
    )
    # compare good and bad guess
    for guess in [bad_guess, good_guess]:
        start = time.time()
        conc_steady = solve(
            methionine.parameters, methionine.unparameterised_model, guess
        )
        jac = jax.jacrev(solve)(
            methionine.parameters, methionine.unparameterised_model, guess
        )
        runtime = (time.time() - start) * 1e3
        sv = model.dcdt(jnp.array(0.0), conc=conc_steady)
        flux = model.flux(conc_steady)
        print(f"Results with starting guess {guess}:")
        print(f"\tRun time in milliseconds: {round(runtime, 4)}")
        print(f"\tSteady state concentration: {conc_steady}")
        print(f"\tFlux: {flux}")
        print(f"\tSv: {sv}")
        print(f"\tJacobian: {jac}")
        print(f"\tLog Km Jacobian: {jac.log_km}")
        print(f"\tDgf Jacobian: {jac.dgf}")


if __name__ == "__main__":
    main()



"""Unit tests for rate equations."""

import pytest
from jax import numpy as jnp
from enzax.rate_equations import (
    AllostericIrreversibleMichaelisMenten,
    AllostericReversibleMichaelisMenten,
    IrreversibleMichaelisMenten,
    ReversibleMichaelisMenten,
)
from enzax.examples import linear


@pytest.mark.parametrize(
    ["rate_equation", "expected_rate"],
    [
        (IrreversibleMichaelisMenten, 0.01105354),
        (ReversibleMichaelisMenten, -0.01188049),
        (AllostericIrreversibleMichaelisMenten, 0.00474257),
        (AllostericReversibleMichaelisMenten, -0.00535258),
    ],
)
def test_rate_equations_linear_one(rate_equation, expected_rate):
    """Check output of different rate equations for rate 1 of linear model."""
    conc = jnp.array([0.5, 0.2, 0.1, 0.3])
    ix = 1
    f = rate_equation(linear.parameters, linear.structure, ix)
    rate = f(conc[linear.structure.rate_to_reactants[ix]])
    assert jnp.isclose(rate, expected_rate)



from abc import abstractmethod, ABC
import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Float, Int, Scalar, ScalarLike, jaxtyped
from typeguard import typechecked


def ragged_jax_index(lol: list[list[int]]) -> list[Int[Array, " _"]]:
    """Convert a list of integer lists into a list of jax int arrays.

    :param lol: a list of lists containing integers.

    """

    def convert_list(loi: list[int]) -> Int[Array, " _"]:
        return jnp.array(loi, dtype=jnp.int16)

    return list(map(convert_list, lol))


@jaxtyped(typechecker=typechecked)
class KineticModelParameters(eqx.Module):
    """Parameters for a kinetic model."""

    log_kcat: Float[Array, " n_enzyme"]
    log_enzyme: Float[Array, " n_enzyme"]
    dgf: Float[Array, " n_metabolite"]
    log_km: Float[Array, " n_km"]
    log_ki: Float[Array, " n_ki"]
    log_conc_unbalanced: Float[Array, " n_unbalanced"]
    temperature: Scalar
    log_transfer_constant: Float[Array, " n_allosteric_enzyme"]
    log_dissociation_constant: Float[Array, " n_allosteric_effector"]
    log_drain: Float[Array, " n_drain"]


@jaxtyped(typechecker=typechecked)
class KineticModelStructure(eqx.Module):
    """Structural information about a kinetic model."""

    S: Float[Array, " s r"] = eqx.field(static=True)
    water_stoichiometry: Float[Array, " r"] = eqx.field(static=True)
    balanced_species: Int[Array, " n_balanced"] = eqx.field(static=True)
    rate_to_enzyme_ix: list[Int[Array, " _"]] = eqx.field(static=True)
    rate_to_km_ixs: list[Int[Array, " _"]] = eqx.field(static=True)
    rate_to_ki_ixs: list[Int[Array, " _"]] = eqx.field(static=True)
    rate_to_tc_ix: list[Int[Array, " _"]] = eqx.field(static=True)
    rate_to_drain_ix: list[Int[Array, " _"]] = eqx.field(static=True)
    drain_sign: Float[Array, " n_drain"] = eqx.field(static=True)
    rate_to_dc_ixs_inhibition: list[Int[Array, " _"]] = eqx.field(static=True)
    rate_to_dc_ixs_activation: list[Int[Array, " _"]] = eqx.field(static=True)
    rate_to_subunits: Int[Array, " r"] = eqx.field(static=True)
    species_to_metabolite_ix: Int[Array, " s"] = eqx.field(static=True)
    ki_to_species_ix: Int[Array, " n_ki"] = eqx.field(static=True)
    dc_to_species_ix: Int[Array, " n_dc"] = eqx.field(static=True)
    unbalanced_species: Int[Array, " n_unbalanced"] = eqx.field(static=True)
    rate_to_reactants: list[Int[Array, " _"]] = eqx.field(static=True)
    rate_to_substrates: list[Int[Array, " _"]] = eqx.field(static=True)
    rate_to_products: list[Int[Array, " _"]] = eqx.field(static=True)
    rate_to_substrate_reactant_positions: list[Int[Array, " _"]] = eqx.field(
        static=True
    )
    rate_to_product_reactant_positions: list[Int[Array, " _"]] = eqx.field(
        static=True
    )
    rate_to_stoichs: list[Float[Array, " _"]] = eqx.field(static=True)

    def __init__(
        self,
        S,
        water_stoichiometry,
        balanced_species,
        rate_to_enzyme_ix,
        rate_to_km_ixs,
        rate_to_ki_ixs,
        rate_to_tc_ix,
        rate_to_drain_ix,
        drain_sign,
        rate_to_dc_ixs_inhibition,
        rate_to_dc_ixs_activation,
        rate_to_subunits,
        species_to_metabolite_ix,
        ki_to_species_ix,
        dc_to_species_ix,
    ):
        self.S = jnp.array(S, dtype=jnp.float64)
        self.water_stoichiometry = jnp.array(
            water_stoichiometry, dtype=jnp.float64
        )
        self.balanced_species = jnp.array(balanced_species, dtype=jnp.int16)
        self.rate_to_enzyme_ix = ragged_jax_index(rate_to_enzyme_ix)
        self.rate_to_km_ixs = ragged_jax_index(rate_to_km_ixs)
        self.rate_to_ki_ixs = ragged_jax_index(rate_to_ki_ixs)
        self.rate_to_tc_ix = ragged_jax_index(rate_to_tc_ix)
        self.rate_to_drain_ix = ragged_jax_index(rate_to_drain_ix)
        self.drain_sign = jnp.array(drain_sign)
        self.rate_to_dc_ixs_inhibition = ragged_jax_index(
            rate_to_dc_ixs_inhibition
        )
        self.rate_to_dc_ixs_activation = ragged_jax_index(
            rate_to_dc_ixs_activation
        )
        self.rate_to_subunits = rate_to_subunits
        self.species_to_metabolite_ix = species_to_metabolite_ix
        self.ki_to_species_ix = ki_to_species_ix
        self.dc_to_species_ix = dc_to_species_ix
        self.unbalanced_species = jnp.array(
            [
                i
                for i in range(self.S.shape[0])
                if i not in self.balanced_species
            ]
        )
        self.rate_to_stoichs = [
            jnp.array(
                [coeff for coeff in coeffs if coeff != 0], dtype=jnp.float64
            )
            for coeffs in self.S.T
        ]
        self.rate_to_reactants = [
            jnp.array(
                [i for i, coeff in enumerate(coeffs) if coeff != 0],
                dtype=jnp.int16,
            )
            for coeffs in self.S.T
        ]
        self.rate_to_substrates = [
            jnp.array(
                [i for i, coeff in enumerate(coeffs) if coeff < 0],
                dtype=jnp.int16,
            )
            for coeffs in self.S.T
        ]
        self.rate_to_products = [
            jnp.array(
                [i for i, coeff in enumerate(coeffs) if coeff > 0],
                dtype=jnp.int16,
            )
            for coeffs in self.S.T
        ]
        self.rate_to_substrate_reactant_positions = [
            jnp.array(
                [i for i, coeff in enumerate(coeffs) if coeff < 0],
                dtype=jnp.int16,
            )
            for coeffs in self.rate_to_stoichs
        ]
        self.rate_to_product_reactant_positions = [
            jnp.array(
                [i for i, coeff in enumerate(coeffs) if coeff > 0],
                dtype=jnp.int16,
            )
            for coeffs in self.rate_to_stoichs
        ]


class RateEquation(ABC, eqx.Module):
    """Class representing an abstract rate equation."""

    @abstractmethod
    def __init__(
        self,
        parameters: KineticModelParameters,
        structure: KineticModelStructure,
        ix: int,
    ):
        """Signature for the __init__ method of a rate equation.

        A rate equation is initialised from a set of parameters, a structure and a positive integer index.
        """
        ...

    @abstractmethod
    def __call__(self, conc: Float[Array, " n"]) -> Scalar:
        """Signature for the __call__ method of a rate equation.

        A rate equation takes in a one dimensional array of positive float-valued concentrations and returns a scalar flux.
        """
        ...


class UnparameterisedKineticModel(eqx.Module):
    """A kinetic model without parameter values."""

    structure: KineticModelStructure
    rate_equation_classes: list[type[RateEquation]]


class KineticModel(eqx.Module):
    """A parameterised kinetic model."""

    parameters: KineticModelParameters
    structure: KineticModelStructure
    rate_equations: list[RateEquation]

    def __init__(self, parameters, unparameterised_model):
        self.parameters = parameters
        self.structure = unparameterised_model.structure
        self.rate_equations = [
            cls(self.parameters, self.structure, ix)
            for ix, cls in enumerate(
                unparameterised_model.rate_equation_classes
            )
        ]

    def flux(
        self, conc_balanced: Float[Array, " n_balanced"]
    ) -> Float[Array, " n"]:
        """Get fluxes from balanced species concentrations.

        :param conc_balanced: a one dimensional array of positive floats representing concentrations of balanced species. Must have same size as self.structure.ix_balanced

        :return: a one dimensional array of (possibly negative) floats representing reaction fluxes. Has same size as number of columns of self.structure.S.

        """
        conc = jnp.zeros(self.structure.S.shape[0])
        conc = conc.at[self.structure.balanced_species].set(conc_balanced)
        conc = conc.at[self.structure.unbalanced_species].set(
            jnp.exp(self.parameters.log_conc_unbalanced)
        )
        t = [f(conc) for ix, f in enumerate(self.rate_equations)]
        out = jnp.array(t)
        return out

    def dcdt(
        self, t: ScalarLike, conc: Float[Array, " n_balanced"], args=None
    ) -> Float[Array, " n_balanced"]:
        """Get the rate of change of balanced species concentrations.

        Note that the signature is as required for a Diffrax vector field function, hence the redundant variable t and the weird name "args".

        """
        out = (self.structure.S @ self.flux(conc))[
            self.structure.balanced_species
        ]
        return out
